/*
 * stormbaancoureur
 * (c) 2007 by Bram Stolk
 * bram at gmail.com
 * LICENSED ACCORDING TO THE GPL
 */

#include <assert.h>
#include <libgen.h>	// for dirname(), Is this portable?

#define GLX_GLXEXT_PROTOTYPES 1

#include <GL/glx.h>
#include <GL/glxext.h>	// for the opengl extensions we use for hardware accelerated shadowing

#include <GL/gl.h>
#include <GL/glut.h>

#include <plib/ssg.h>
#include <plib/ul.h>

#include "staticworldobject.h"
#include "carobject.h"
#include "trackingcam.h"
#include "usercam.h"
#include "dynamicboxobject.h"
#include "dynamiccylobject.h"

#include "stereocontext.h"
#include "intro.h"
#include "controller.h"
#include "soundenginealsa.h"
#include "plodegui.h"
#include "modelmap.h"
#include "postscore.h"
#include "sturmbahn.h"
#include "starsky.h"
#include "texture.h"

static float aspectratio=1.0;
//static int winw=704, winh=300;
static int winw=480, winh=360;
static ssgRoot *scene=0;
static ssgContext    *monocontext=0;
static StereoContext *stereocontext=0;
static ssgContext    *lightviewcontext=0;
static float fps=60.0;
static float gametime=0.0;
static char *username="anonymou";
static float dt_hist[10]={0.016,0.016,0.016,0.016,0.016,0.016,0.016,0.016,0.016,0.016};
static std::string dirprefix;

static dJointGroupID contactgroup;
static dSpaceID bigspace;
static dSpaceID staticspace;
static dWorldID world;

static SturmBahnFull     *sturmbahnfull=0;
static SturmBahnPractice *sturmbahnpractice=0;
static SturmBahnRally    *sturmbahnrally=0;
static RespawnPoint      *respawn_point_hit=0;

static ControllerPad *controller_pad=0;
static ControllerKey *controller_key=0;
static Controller    *controller=0;
static CarObject *car=0;
//static TrackingCam *trackingcam=0;
static UserCam *trackingcam=0;
static int numcars=0;

static SoundEngineAlsa *sengine=0;
static Intro *intro=0;
static PlodeGUI *gui=0;
static bool capture=false;

static ModelMap *modelmap=0;
static StarSky  *starsky=0;

static float camd=1.8;
static float cama=1.25*M_PI;

// Variables used for hardware accelerated shadow mapping
static bool do_shadows;
static GLuint frame_buffer_object;
static GLuint light_view_depth_texture;
static Texture *spotlight_texture=0;
static const int shadowmapsize = 1024;
static const float spot_fov = 50.0f;
static GLuint fragment_shader, vertex_shader;
static GLuint program;
static GLint uniform_loc_tex0, uniform_loc_tex1, uniform_loc_tex2;


static void stop_game(void);
static void start_game(const std::string &gametype);
static bool respawn(void);



static void OglErrorCheck(const std::string &context)
{
  GLenum error = glGetError();
  while (error != GL_NO_ERROR)
  {
    fprintf(stderr, "plode OpenGL error (%s): %s\n", context.c_str(),gluErrorString(error));
    error = glGetError();
  }
}


/*
 * This helper function checks wether we have all req'd extensions for shadow mapping
 */
static bool OglCanDoShadowing(void)
{
  if (getenv("PLODE_NO_SHADOWS")) return false;
  const char *vers = (const char*) glGetString(GL_VERSION);
  const char *rend = (const char*) glGetString(GL_RENDERER);
  fprintf(stderr, "OpenGL is version %s\n", vers);
  fprintf(stderr, "OpenGL renderer %s\n", rend);
  const char *exts = (const char*) glGetString(GL_EXTENSIONS);
  //fprintf(stderr, "OpenGL has extensions %s\n", exts);
  if (!strstr(exts, "GL_ARB_multitexture")) return false;
  if (!strstr(exts, "GL_ARB_depth_texture")) return false;
  if (!strstr(exts, "GL_ARB_shadow")) return false;
  if (!strstr(exts, "GL_EXT_framebuffer_object")) return false;
  if (!strstr(exts, "GL_ARB_vertex_shader")) return false;
  if (!strstr(exts, "GL_ARB_fragment_shader")) return false;
  if (!strstr(exts, "GL_ARB_shader_objects")) return false;
  return true;
}


static void OglPrepareFragmentShader(void)
{
  PFNGLCREATESHADERPROC  glCreateShader  = (PFNGLCREATESHADERPROC)  glXGetProcAddressARB((const GLubyte *) "glCreateShader");
  PFNGLSHADERSOURCEPROC  glShaderSource  = (PFNGLSHADERSOURCEPROC)  glXGetProcAddressARB((const GLubyte *) "glShaderSource");
  PFNGLCOMPILESHADERPROC glCompileShader = (PFNGLCOMPILESHADERPROC) glXGetProcAddressARB((const GLubyte *) "glCompileShader");
  PFNGLGETSHADERIVPROC   glGetShaderiv   = (PFNGLGETSHADERIVPROC)   glXGetProcAddressARB((const GLubyte *) "glGetShaderiv");
  PFNGLCREATEPROGRAMPROC glCreateProgram = (PFNGLCREATEPROGRAMPROC) glXGetProcAddressARB((const GLubyte *) "glCreateProgram");
  PFNGLATTACHSHADERPROC  glAttachShader  = (PFNGLATTACHSHADERPROC)  glXGetProcAddressARB((const GLubyte *) "glAttachShader");
  PFNGLLINKPROGRAMPROC   glLinkProgram   = (PFNGLLINKPROGRAMPROC)   glXGetProcAddressARB((const GLubyte *) "glLinkProgram");
  PFNGLGETPROGRAMIVPROC  glGetProgramiv  = (PFNGLGETPROGRAMIVPROC)  glXGetProcAddressARB((const GLubyte *) "glGetProgramiv");
  PFNGLGETSHADERINFOLOGPROC glGetShaderInfoLog = (PFNGLGETSHADERINFOLOGPROC) glXGetProcAddressARB((const GLubyte *) "glGetShaderInfoLog");
  PFNGLGETPROGRAMINFOLOGPROC glGetProgramInfoLog = (PFNGLGETPROGRAMINFOLOGPROC) glXGetProcAddressARB((const GLubyte *) "glGetProgramInfoLog");
  PFNGLGETUNIFORMLOCATIONPROC glGetUniformLocation = (PFNGLGETUNIFORMLOCATIONPROC) glXGetProcAddressARB((const GLubyte *) "glGetUniformLocation");

  assert(glCreateShader);
  assert(glShaderSource);
  assert(glCompileShader);
  assert(glGetShaderiv);
  assert(glCreateProgram);
  assert(glLinkProgram);
  assert(glAttachShader);
  assert(glGetProgramiv);
  assert(glGetShaderInfoLog);
  assert(glGetUniformLocation);

  static GLchar fragment_program_text[8192];
  static GLchar vertex_program_text[8192];
  const char *fp = fragment_program_text;
  const char *vp = vertex_program_text;
  char log_text[4096];
  int sz;
  FILE *f;
  std::string fname;
  GLint status;

  fname = dirprefix + "/shaders/bramlight.fp";
  f = fopen(fname.c_str(),"rb"); assert(f); sz=fread(fragment_program_text, 1, 8192, f); fclose(f);
  sz = strlen(fragment_program_text);
  fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(fragment_shader, 1, &fp, 0);
  glCompileShader(fragment_shader);
  glGetShaderiv(fragment_shader, GL_COMPILE_STATUS, &status);
  if (status != GL_TRUE)
  {
    glGetShaderInfoLog(fragment_shader, sizeof(log_text), 0, log_text);
    fprintf(stderr,"%s\n",log_text);
    exit(1);
  }

  fname = dirprefix + "/shaders/bramlight.vp";
  f = fopen(fname.c_str(),"rb"); assert(f); sz=fread(vertex_program_text, 1, 8192, f); fclose(f);
  sz = strlen(vertex_program_text);
  vertex_shader = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(vertex_shader, 1, &vp, 0);
  glCompileShader(vertex_shader);
  glGetShaderiv(vertex_shader, GL_COMPILE_STATUS, &status);
  if (status != GL_TRUE)
  {
    glGetShaderInfoLog(vertex_shader, sizeof(log_text), 0, log_text);
    fprintf(stderr,"%s\n",log_text);
    exit(1);
  }

  program = glCreateProgram();
  glAttachShader(program, vertex_shader);
  glAttachShader(program, fragment_shader);
  glLinkProgram(program);
  glGetProgramiv(program, GL_LINK_STATUS, &status);
  if (status != GL_TRUE)
  {
    glGetProgramInfoLog(program, sizeof(log_text), 0, log_text);
    fprintf(stderr,"%s\n",log_text);
    exit(1);
  }

  // Find the locations of the uniforms that we are using.
  uniform_loc_tex0 = glGetUniformLocation(program, "texture0");
  uniform_loc_tex1 = glGetUniformLocation(program, "texture1");
  uniform_loc_tex2 = glGetUniformLocation(program, "texture2");
  assert(uniform_loc_tex0 != -1);
  assert(uniform_loc_tex1 != -1);
  assert(uniform_loc_tex2 != -1);

  OglErrorCheck("prepare fragment shader");
}


/*
 * This piece of code sets up a frame buffer object, associated with a texture.
 * We can render a depth-only image as seen from the light, into this texture.
 */
static void OglInitShadowing(void)
{
  // create texture for shadow map
  glGenTextures(1, &light_view_depth_texture);
  glBindTexture(GL_TEXTURE_2D, light_view_depth_texture);
  glEnable(GL_DEPTH_TEST);
  GLint bits;
  glGetIntegerv(GL_DEPTH_BITS, &bits);
  glTexImage2D
  (
    GL_TEXTURE_2D,
    0,
    (bits==16) ? GL_DEPTH_COMPONENT16_ARB : GL_DEPTH_COMPONENT24_ARB,
    shadowmapsize, shadowmapsize, 0,
    GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, 0
  );
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
  OglErrorCheck("creating shadow texture");

  // create texture for spotlight image
  spotlight_texture = new Texture(dirprefix+"/images/spot.tga");
  spotlight_texture->InitForRendering();
  OglErrorCheck("creating spotlight texture");

  // Get the addresses of the extended functionality.
  PFNGLGENFRAMEBUFFERSEXTPROC glGenFramebuffers = (PFNGLGENFRAMEBUFFERSEXTPROC) glXGetProcAddressARB((const GLubyte *) "glGenFramebuffersEXT");
  PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebuffer = (PFNGLBINDFRAMEBUFFEREXTPROC) glXGetProcAddressARB((const GLubyte *) "glBindFramebufferEXT");
  PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2D = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC) glXGetProcAddressARB((const GLubyte *) "glFramebufferTexture2DEXT");
  PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatus = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC) glXGetProcAddressARB((const GLubyte *) "glCheckFramebufferStatusEXT");

  assert(glGenFramebuffers);
  assert(glBindFramebuffer);
  assert(glFramebufferTexture2D);
  assert(glCheckFramebufferStatus);

  // create fbo
  glGenFramebuffers(1, &frame_buffer_object);
  glBindFramebuffer(GL_FRAMEBUFFER_EXT, frame_buffer_object);
  // Attach texture to fbo
  glFramebufferTexture2D
  (
    GL_FRAMEBUFFER_EXT,
    GL_DEPTH_ATTACHMENT_EXT,
    GL_TEXTURE_2D,
    light_view_depth_texture,
   0
  );

  // We use depth render only
  glDrawBuffer(GL_NONE);
  glReadBuffer(GL_NONE);

  GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER_EXT);
  assert(status != GL_FRAMEBUFFER_UNSUPPORTED_EXT);
  //fprintf(stderr,"status %x\n", status);
  assert(status == GL_FRAMEBUFFER_COMPLETE_EXT);

  // disable the frame buffer object
  glBindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
  OglErrorCheck("creating frame buffer object");

  OglPrepareFragmentShader();
}


static void set_linear_texgen(GLenum plane, GLenum mode)
{
  // eye linear
  GLfloat xx[4]={1,0,0,0};
  GLfloat yy[4]={0,1,0,0};
  GLfloat zz[4]={0,0,1,0};
  GLfloat ww[4]={0,0,0,1};
  glTexGenfv(GL_S, plane, xx);
  glTexGenfv(GL_T, plane, yy);
  glTexGenfv(GL_R, plane, zz);
  glTexGenfv(GL_Q, plane, ww);
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, mode);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, mode);
  glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, mode);
  glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, mode);
  glEnable(GL_TEXTURE_GEN_S);
  glEnable(GL_TEXTURE_GEN_T);
  glEnable(GL_TEXTURE_GEN_R);
  glEnable(GL_TEXTURE_GEN_Q);
}


static void render_shadow_map(void)
{
  // Prepare shadow rendering
  int port[4];
  glGetIntegerv(GL_VIEWPORT, port);
  PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebuffer = (PFNGLBINDFRAMEBUFFEREXTPROC) glXGetProcAddressARB((const GLubyte *) "glBindFramebufferEXT");
  glBindFramebuffer(GL_FRAMEBUFFER_EXT, frame_buffer_object);
  glViewport(0,0, shadowmapsize,shadowmapsize);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // depth only?
//  glPolygonOffset(2.5, 10);
  glPolygonOffset(1.1, 4.0);
  glEnable(GL_POLYGON_OFFSET_FILL);
  OglErrorCheck("prepare shadow rendering");

  // render scene from light view

  glActiveTextureARB(GL_TEXTURE1_ARB);

  glPushMatrix();
  set_linear_texgen(GL_EYE_PLANE, GL_EYE_LINEAR);
  glPopMatrix();

  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  glTranslatef(0.5f, 0.5f, 0.5f);
  glScalef(0.5f, 0.5f, 0.5f);
  gluPerspective(spot_fov, 1, 0.1, 150.0);
  glMatrixMode(GL_MODELVIEW);

  spotlight_texture->Activate();
  glEnable(GL_TEXTURE_2D);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  // Render the scene
  glActiveTextureARB(GL_TEXTURE0_ARB);
  set_linear_texgen(GL_OBJECT_PLANE, GL_OBJECT_LINEAR);
  ssgCullAndDraw(scene);
  OglErrorCheck("ssgCullAndDraw");

  glActiveTextureARB(GL_TEXTURE1_ARB);
  glBindTexture(GL_TEXTURE_2D, 0);
  glDisable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE0_ARB);
  glBindTexture(GL_TEXTURE_2D, 0);
  glDisable(GL_TEXTURE_2D);

  // reset viewport and polygon offset to original values.
  glDisable(GL_POLYGON_OFFSET_FILL);
  glBindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
  glViewport(port[0], port[1], port[2], port[3]);
  OglErrorCheck("render_shadow_map");
}


static void near_callback(void *data, dGeomID o1, dGeomID o2)
{
  assert(o1);
  assert(o2);

  if (dGeomIsSpace(o1) || dGeomIsSpace(o2))
  {
    // colliding a space with something
    dSpaceCollide2(o1,o2,data,&near_callback);
    // Note we do not want to test intersections within a space,
    // only between spaces.
    return;
  }

  // Two non space geoms

  WorldObject *wo1 = static_cast<WorldObject*>(dGeomGetData(o1));
  WorldObject *wo2 = static_cast<WorldObject*>(dGeomGetData(o2));

  const int N = 32;
  dContact contact[N];
  int n = dCollide (o1,o2,N,&(contact[0].geom),sizeof(dContact));

  if (n > 0)
  {
    assert(wo1);
    assert(wo2);

    if (wo1->name == "respawnpoint")
      respawn_point_hit = dynamic_cast<RespawnPoint*>(wo1);
    if (wo2->name == "respawnpoint")
      respawn_point_hit = dynamic_cast<RespawnPoint*>(wo2);

    // This section is req'd to see if the backwheels are touching
    // a dynamic object. If so, it could be that our car is riding
    // on top of an object. In this case, it should be possible to
    // engage the reverse-gear when the car is stationary relative
    // to the ground, and yet may be moving in absolute space.
    if (car->IsBackWheelGeometry(o1))
    {
      DynamicObject *dynamic_ground = dynamic_cast<DynamicObject*>(wo2);
      if (dynamic_ground)
        car->SetGround(dynamic_ground);
      else
        car->SetGround(0);
    }
    if (car->IsBackWheelGeometry(o2))
    {
      DynamicObject *dynamic_ground = dynamic_cast<DynamicObject*>(wo1);
      if (dynamic_ground)
        car->SetGround(dynamic_ground);
      else
        car->SetGround(0);
    }

    bool is_tyre=false;
    sgVec3 wheel_dir={0,0,0};
    float wheel_vel=0.0f;
    if (car->IsWheelGeometry(o1,wheel_vel))
    {
      const dReal *m = dGeomGetRotation(o1);
      sgSetVec3(wheel_dir, m[2],m[6],m[10]);
      is_tyre=true;
    }
    if (car->IsWheelGeometry(o2,wheel_vel))
    {
      const dReal *m = dGeomGetRotation(o2);
      sgSetVec3(wheel_dir, m[2],m[6],m[10]);
      is_tyre=true;
    }

#if 0
    if (is_tyre)
      fprintf(stderr,"dir %5.2f %5.2f %5.2f\n", wheel_dir[0], wheel_dir[1], wheel_dir[2]);
#endif

    // Create constraints for all the contacts between the geometries.
    for (int i=0; i<n; i++)
    {
      contact[i].fdir1[0]=wheel_dir[0];
      contact[i].fdir1[1]=wheel_dir[1];
      contact[i].fdir1[2]=wheel_dir[2];
      contact[i].surface.slip1 = 0.04f * wheel_vel; // slip perpendicular to roll direction
      contact[i].surface.slip2 = 0.02f; // slip in roll direction
      contact[i].surface.mu = 50.0;
      contact[i].surface.soft_erp = 0.98;
      contact[i].surface.soft_cfm = 0.03;

      contact[i].surface.mode = dContactSoftERP | dContactSoftCFM | dContactApprox1;
      if (is_tyre) 
      {
        contact[i].surface.mode = contact[i].surface.mode | dContactFDir1 | dContactSlip1 | dContactSlip2;
        contact[i].surface.mu = 400.0;
        contact[i].surface.soft_erp = 0.95;
        contact[i].surface.soft_cfm = 0.05;
      }

      if ((wo1->name == "ferriswheelcart" || wo2->name == "ferriswheelcart") && is_tyre)
      {
        contact[i].surface.mu = 700.0;
        contact[i].surface.slip1 *= 0.01;
        contact[i].surface.slip2 *= 0.01;
      }

      dJointID c = dJointCreateContact (world,contactgroup,&contact[i]);
      dGeomID g1 = contact[i].geom.g1;
      dGeomID g2 = contact[i].geom.g2;
      dBodyID b1 = dGeomGetBody(g1);
      dBodyID b2 = dGeomGetBody(g2);
      assert(b1 || b2);
      dJointAttach (c, b1, b2);
    }
  }
}


static void engine_noise(float dt)
{
  float vl = car->GetWheelVelocity(2);
  float vr = car->GetWheelVelocity(3);
  float v = std::max(vl,vr);
  float a = car->GetAccelerator();
  float d = car->GetBrake();
  float e = std::max(a,d);
  v = v/15.0;
  if (v>3) v=3;
  sengine->SetModulation(1+v+e, 1+v/5+e/2);
  sengine->SetLowPass(0.97);
  return;
}


void make_menu_choice(void)
{
  const std::string &choice = gui->SelectChoice();
  if (choice == "ATTACK STORMBAAN")
    start_game("full");
  if (choice == "PRACTICE")
    start_game("practice");
  if (choice == "TEST GROUND")
    start_game("rally");
}


static void idle(void)
{
  static ulClock clk;
  static int frameCounter = 0;
  clk.setMaxDelta(0.4);  // Don't accomodate for systems slower than 2.5 fps

  float elapsed = clk.getDeltaTime();
  clk.update();

  float fractiondone=0;
  if (sengine)
    fractiondone=sengine->Sustain();
  if (intro)
  {
    bool done=intro->Sustain(elapsed, fractiondone);
    if (done || !gui->hasmenu)
    {
      delete intro;
      intro = 0;
    }
  }

  dt_hist[frameCounter] = elapsed;
  frameCounter = (frameCounter+1)%10;

  float dt=0;
  int i;
  for (i=0; i<10; i++)
    dt += dt_hist[i];
  dt = dt / 10;
  assert(dt>0.0);

  if (controller_pad) controller_pad->Sustain(dt);
  if (controller_key) controller_key->Sustain(dt);

  // Feed the controls into the car
  float ste = controller->GetSteer();
  float acc = controller->GetAccel();
  float bra = controller->GetBrake();
  bool  act = controller->GetAction();
  bool  ebr = controller->GetEBrake();

  // Steer the gui with the gamepad
  if (gui && gui->hasmenu && controller_pad)
  {
    if (controller_pad->HasAnyButtonChanged(true))
      make_menu_choice();
    static float delta=0.0;
    delta += controller_pad->GetAxisValue(1)*10*dt;
    int idelta = int(delta);
    if (idelta)
    {
      delta=0.0;
      gui->ChangeChoice(idelta);
    }
  }

  // do not update the fps every time, it is not readable when it is done each tick
  if (frameCounter == 0) 
    fps = 1.0f / dt;

  if (intro)
    return;

  if (gui && gui->hasmenu)
    return;

  car->SetSteer(ste);
  car->SetAccelerator(acc);
  car->SetBrake(bra);
  car->SetAction(act);
  car->SetEBrake(ebr);

  gametime += elapsed;
  if (sengine)
    engine_noise(dt);

  float timestep=0.008333; // Use 8.333ms timesteps
  static float remaining_sim_time=0;
  remaining_sim_time += dt;

  bool win=false;
  const float SLOWMOTION=1.0f;
  while (remaining_sim_time > timestep)
  {
    dSpaceCollide (bigspace, 0, &near_callback);
    if (car) car->Sustain(SLOWMOTION*timestep);
    if (sturmbahnpractice)
      sturmbahnpractice->Sustain(SLOWMOTION*timestep, car->GetAction());
    if (sturmbahnfull)
      if (sturmbahnfull->Sustain(SLOWMOTION*timestep, car->GetAction()))
        win=true;
    if (sturmbahnrally)
      sturmbahnrally->Sustain(SLOWMOTION*timestep, car->GetAction());
    dWorldQuickStep (world, SLOWMOTION*timestep);
    dJointGroupEmpty (contactgroup);
    remaining_sim_time -= timestep;
  }

  if (win)
  {
    postscore_put(username, gametime);
    const char *leaderboardtext = postscore_get();
    gui->SetLeaderBoardText(leaderboardtext);
    stop_game();
    return;
  }

  // Check lose condition
  sgVec3 carpos;
  car->GetPos(carpos);
  if (carpos[2] < -4)
  {
    const char *gametype="";
    if (sturmbahnpractice) gametype="practice";
    if (sturmbahnrally) gametype="rally";
    if (sturmbahnfull) gametype="full";
    bool respawned = respawn();
    if (!respawned)
    {
      stop_game();
      start_game(gametype);
    }
    return;
  }

  if (respawn_point_hit)
  {
    car->SnapShot(false);
    gametime += 60.0f; // penalty for taking the easy way out
    respawn_point_hit->Disable();
    respawn_point_hit=0;
  }

  trackingcam->Update(dt);
  sgVec3 eye, coi, up;
  sgSetVec3(up,0,0,1);
  trackingcam->GetCameraPos(eye);
  trackingcam->GetTargetPos(coi);

  if (monocontext)
    monocontext->setCameraLookAt(eye, coi, up);

  if (stereocontext)
  {
    stereocontext->SetCameraLookAt(eye, coi, up);
    OglErrorCheck("stereocontext->SetCameraLookAt");
  }

  ssgLight *light = ssgGetLight(0);
  car->GetPos(coi);
  car->GetPos(eye);
  eye[2] += 6.5;
  sgVec3 car_dir;
  car->GetDir(car_dir);
  sgAddVec3(coi, coi, car_dir);
  sgVec3 light_dir;
  sgSubVec3(light_dir, coi, eye);
  light->setPosition(eye[0],eye[1],eye[2]);
  light->setSpotDirection(light_dir[0],light_dir[1],light_dir[2]);
  sgSetVec3(up,1,0,0);
  lightviewcontext->setCameraLookAt(eye, coi, up);
}


static void redraw_single_channel(void)
{
  if (scene && !intro && (!gui || !gui->hasmenu))
  {
    static PFNGLUSEPROGRAMPROC glUseProgram = (PFNGLUSEPROGRAMPROC) glXGetProcAddressARB((const GLubyte *) "glUseProgram");
    static PFNGLUNIFORM1IPROC  glUniform1i  = (PFNGLUNIFORM1IPROC)  glXGetProcAddressARB((const GLubyte *) "glUniform1i");

    if (do_shadows)
    {
      assert(glUseProgram);
      // 1st pass from light view

      lightviewcontext->makeCurrent();
      render_shadow_map();

      sgMat4 mat_cam, mat_spt;
      if (monocontext)
      {  
        monocontext->makeCurrent();
        monocontext->getModelviewMatrix(mat_cam);
      }
      if (stereocontext) 
      {
        stereocontext->MakeCurrent("last");
        stereocontext->getModelviewMatrix(mat_cam);
      }
      lightviewcontext->getModelviewMatrix(mat_spt);

      // 2nd pass: with shadow texture
  
      glUseProgram(program);
      glActiveTextureARB(GL_TEXTURE1_ARB);
      glUniform1i(uniform_loc_tex1, 1);
      glPushMatrix();
      glMultMatrixf((GLfloat*)mat_cam);
      set_linear_texgen(GL_EYE_PLANE, GL_EYE_LINEAR);
      glPopMatrix();
      glMatrixMode(GL_TEXTURE);
      glLoadIdentity();
      glTranslatef(0.5f,0.5f,0.5f);
      glScalef(0.5f,0.5f,0.5f);
      gluPerspective(spot_fov, 1, 1.0, 150.0);
      glMultMatrixf((GLfloat*)mat_spt);
      glMatrixMode(GL_MODELVIEW);
      spotlight_texture->Activate();
      glEnable(GL_TEXTURE_2D);
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

      glActiveTextureARB(GL_TEXTURE2_ARB);
      glUniform1i(uniform_loc_tex2, 2);
      glPushMatrix();
      glMultMatrixf((GLfloat*)mat_cam);
      set_linear_texgen(GL_EYE_PLANE, GL_EYE_LINEAR);
      glPopMatrix();
      glMatrixMode(GL_TEXTURE);
      glLoadIdentity();
      glTranslatef(0.5f, 0.5f, 0.5f); // r coord scale
      glScalef(0.5f, 0.5f, 0.5f); // r coord bias
      gluPerspective(spot_fov, 1, 1.0f, 150.0f);
      glMultMatrixf((GLfloat*)mat_spt);
      OglErrorCheck("setup texture matrix");

      glMatrixMode(GL_MODELVIEW);
      glBindTexture(GL_TEXTURE_2D, light_view_depth_texture);
      glEnable(GL_TEXTURE_2D);
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
 
      glActiveTextureARB(GL_TEXTURE0_ARB);
      glUniform1i(uniform_loc_tex0, 0);
      set_linear_texgen(GL_OBJECT_PLANE, GL_OBJECT_LINEAR);
    }

    ssgCullAndDraw(scene);
    OglErrorCheck("ssgCullAndDraw");

    if (do_shadows)
    {
      glUseProgram(0);
      // Turn off all texture units
      glActiveTextureARB(GL_TEXTURE1_ARB);
      glBindTexture(GL_TEXTURE_2D, 0);
      glDisable(GL_TEXTURE_2D);
      glActiveTextureARB(GL_TEXTURE2_ARB);
      glBindTexture(GL_TEXTURE_2D, 0);
      glDisable(GL_TEXTURE_2D);
      glActiveTextureARB(GL_TEXTURE0_ARB);
      glBindTexture(GL_TEXTURE_2D, 0);
      glDisable(GL_TEXTURE_2D);
      OglErrorCheck("Turn off texture units");
    }
  }

  if (intro)
  {
    intro->Redraw(winw, winh);
    OglErrorCheck("intro->Redraw");
  }

  if (gui)
  {
    char s[25];
    sprintf(s, "%.02f   %d FPS", gametime, (int) fps);
    gui->Redraw(winw, winh, s);
    OglErrorCheck("gui->Redraw");
  }
}


static void redraw(void)
{
  if (intro || (gui && gui->hasmenu))
    glClearColor(0.3,0.3,0.55,1);
  else
    glClearColor(0.2,0.1,0.35,1);

  if (stereocontext)
  {
    // passive stereo: clear buffers only once
    if (!stereocontext->IsQuadBuffered())
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
    stereocontext->MakeCurrent("left");
    OglErrorCheck("MakeCurrent");
#if 1
    // quad buffer stereo: clear buffers twice
    if (stereocontext->IsQuadBuffered())
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
#endif
    redraw_single_channel();
    stereocontext->MakeCurrent("right");
    OglErrorCheck("MakeCurrent");
#if 1
    // quad buffer stereo: clear buffers twice
    if (stereocontext->IsQuadBuffered())
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
#endif
    redraw_single_channel();
  }

  if (monocontext)
  {
    monocontext->makeCurrent();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ) ;
    redraw_single_channel();
  }

  glutSwapBuffers () ;
  OglErrorCheck("glutSwapBuffers");

  if (capture)
  {
    static char *buf = new char[winw * winh * 3];
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glReadBuffer(GL_BACK_LEFT);
    glReadPixels(0,0,winw,winh,GL_RGB,GL_UNSIGNED_BYTE,(GLvoid*)(buf));
    OglErrorCheck("glReadPixels");
    static FILE *f=0;

    if (!f)
    {
      f=fopen("capture.out", "wb");
      assert(f);
 //        fprintf(f,"P6 %d %d 255\n", winw,winh);
    }
    unsigned char *revbuf = new unsigned char[winw*winh*3];
    for (int y=0; y<winh; y++)
    {
      int chunk = 3 * winw;
      memcpy(revbuf + y*chunk, buf + (winh - 1 - y)*chunk, chunk);
    }
    fwrite(revbuf,winw*winh*3,1,f);
    fflush(f);
    delete revbuf;
  }
  // We want to render the next frame immediately
  glutPostRedisplay () ;
}


static void reshape(int w, int h)
{
  winw=w; winh=h;
  if (monocontext)
  {
    glViewport ( 0, 0, w, h ) ;
    aspectratio = w / (float) h;
//    float fovy = 60.0 * M_PI / 180.0;
    float fovy = 40.0 * M_PI / 180.0;
    float nearPlaneDistance = 0.28;
    float farPlaneDistance  = 2500.0;
    float y = tan(0.5 * fovy) * nearPlaneDistance;
    float x = aspectratio * y;
    monocontext->setFrustum(-x,x,-y,y,nearPlaneDistance,farPlaneDistance);
  }

  if (stereocontext)
    stereocontext->SetWindowSize(winw, winh);

#if 1
  int nsamples = glutGet(GLUT_WINDOW_NUM_SAMPLES);
#else
  int nsamples;
  glGetIntegerv(GL_SAMPLES, &nsamples);
#endif
  fprintf(stderr,"Number of samples per pixel: %d\n", nsamples);
}


static void keyboardUp(unsigned char k, int, int)
{
  if (controller_key) controller_key->FeedKey(k, false);
}


static void special(int k, int, int)
{
  if (k == GLUT_KEY_UP)
  {
    if (gui) gui->ChangeChoice(-1);
    if (controller_key) controller_key->FeedAction("ACCEL", true);
  }
  if (k == GLUT_KEY_DOWN)
  {
    if (gui) gui->ChangeChoice(1);
    if (controller_key) controller_key->FeedAction("BRAKE", true);
  }
  if (k == GLUT_KEY_LEFT)
  {
    if (controller_key) controller_key->FeedAction("LEFT", true);
  }
  if (k == GLUT_KEY_RIGHT)
  {
     if (controller_key) controller_key->FeedAction("RIGHT", true);
  }
}


static void specialUp(int k, int, int)
{
  if (k == GLUT_KEY_UP)
  {
     if (controller_key) controller_key->FeedAction("ACCEL", false);
  }
  if (k == GLUT_KEY_DOWN)
  {
    if (controller_key) controller_key->FeedAction("BRAKE", false);
  }
  if (k == GLUT_KEY_LEFT)
  {
    if (controller_key) controller_key->FeedAction("LEFT", false);
  }
  if (k == GLUT_KEY_RIGHT)
  {
    if (controller_key) controller_key->FeedAction("RIGHT", false);
  }
}


static int mouse_x, mouse_y, mouse_but;
static void mouse(int button, int state, int x, int y)
{
  mouse_x = x;
  mouse_y = y;
  if (state == GLUT_DOWN) mouse_but = button;
}


static void motion(int x, int y)
{
  if (!trackingcam) return;
  float dx = (x - mouse_x) / (float) winw;
  float dy = (y - mouse_y) / (float) winh;

  if (mouse_but == GLUT_LEFT_BUTTON)
  {
    trackingcam->ChangeHeading(100.0 * dx);
    trackingcam->ChangePitch(100.0 * dy);
  }
  if (mouse_but == GLUT_MIDDLE_BUTTON)
  {
    trackingcam->ChangeOffset(dx, dy);
  }
  if (mouse_but == GLUT_RIGHT_BUTTON)
  {
    trackingcam->ChangeDistance(-10.0 * dy);
  }

  mouse_x = x;
  mouse_y = y;
}


static void start_game(const std::string &gametype)
{
  gametime = 0.0;

  // Create ODE world
  dInitODE();
  world = dWorldCreate();
  bigspace = dHashSpaceCreate(0);
  staticspace = dSimpleSpaceCreate(bigspace);
  contactgroup = dJointGroupCreate (0);
  dWorldSetGravity(world,0,0,-9.8);
  dWorldSetAutoDisableFlag(world, true);
  dWorldSetAutoDisableLinearThreshold(world, 0.04);
  dWorldSetAutoDisableAngularThreshold(world, 0.04);
  dWorldSetQuickStepNumIterations(world, 64);

  // Create PLIB world

  scene = new ssgRoot();

  if (gametype=="practice")
    sturmbahnpractice = new SturmBahnPractice(world, bigspace, staticspace, scene, modelmap, dirprefix);
  if (gametype=="rally")
    sturmbahnrally = new SturmBahnRally(world, bigspace, staticspace, scene, modelmap, dirprefix);
  if (gametype=="full")
    sturmbahnfull = new SturmBahnFull(world, bigspace, staticspace, scene, modelmap, dirprefix);

  // Create a car
  sgVec3 carpos;
  sgSetVec3(carpos, -4, 0, 1.5);

  //sgSetVec3(carpos, 52, 0, 1.5);  // If you wanna cheat!

  if (sturmbahnrally)
    sgSetVec3(carpos, -24, 0, 1.5);

  ssgEntity *licplatemodel = 0;
#if 0
  if (!modelmap->LowDetail())
    licplatemodel = modelmap->Get("licplate.ac");
#endif
  car = new CarObject
  (
    modelmap->Get("frame.3ds"),
    modelmap->Get("fivespoke.3ds"),
    modelmap->Get("wishbone.3ds"),
    modelmap->Get("spindle.3ds"),
    modelmap->Get("coilspring.3ds"),
    modelmap->Get("leafspring.3ds"),
    modelmap->Get("rearaxle.3ds"),
    licplatemodel,
    world,
    bigspace,
    carpos
  );
  scene->addKid(car->GetEntity());
  numcars = 1;

  // Setup camera
  trackingcam = new UserCam(0,0,0);
  trackingcam->AddTarget(car->GetTransform());
  trackingcam->SetPitch(18.0f);
  trackingcam->SetDistance(2.2f);

  if (sengine)
  {
    sengine->SetMode("engine");
    sengine->Play(dirprefix + "/sounds/detonationnorm_s16_le.wav",0);
  }
}


static void stop_game(void)
{
  delete car;
  delete trackingcam;

  if (sengine)
    sengine->SetMode("none");

  delete sturmbahnfull;
  sturmbahnfull=0;
  delete sturmbahnpractice;
  sturmbahnpractice=0;
  delete sturmbahnrally;
  sturmbahnrally=0;

  // Delete ODE stuff
  dJointGroupDestroy(contactgroup);
  dSpaceDestroy(staticspace);
  dSpaceDestroy(bigspace);
  dWorldDestroy(world);

  dCloseODE();
}


static bool respawn(void)
{
  bool restored = car->SnapShot(true);
  if (!restored)
    return false;
  return true;
}


static void keyboard(unsigned char k, int, int)
{
  if (k==27)
  {
    if (gui && gui->hasmenu)
    {
      if (gui->showingleaders)
      {
        gui->showingleaders = false;
        return;
      }
      if (gui->showinghelp)
      {
        gui->showinghelp = false;
        return;
      }
      exit(0);
    }
    stop_game();
    gametime=0.0;
    if (gui) gui->hasmenu=true;
  }

  if (gui && gui->hasmenu)
  {
    if (k==' ' || k==13 || k==10)
    {
      make_menu_choice();
    }
    return;
  }

  static float yoff=0.0;
  static bool autocam=false;
  if (k=='c')
  {
    autocam = !autocam;
//    trackingcam->SetAutoCam(autocam);
    if (!autocam)
    {
//      trackingcam->SetTiltAngle(cama);
//      trackingcam->SetDistance(camd);
    }
  }
  if (k=='2' && !autocam)
  {
    cama = 1.1 * cama;
//    trackingcam->SetTiltAngle(cama);
    return;
  }
  if (k=='1' && !autocam)
  {
    cama = 0.9 * cama;
//    trackingcam->SetTiltAngle(cama);
    return;
  }
  if (k=='3')
  { 
    yoff--;
// trackingcam->SetYOff(yoff);
    return;
  }
  if (k=='4')
  { 
    yoff++;
//    trackingcam->SetYOff(yoff);
    return;
  }
  if (k=='=')
  {
    camd = 1.1 * camd;
//    trackingcam->SetDistance(camd);
    return;
  }
  if (k=='-')
  {
    camd = 0.9 * camd;
//    trackingcam->SetDistance(camd);
    return;
  }
  if (k=='s')
  {
    car->SnapShot(true);
    return;
  }
#if 1
  if (k=='x')
  {
    capture=!capture;
    return;
  }
#endif

  if (controller_key) controller_key->FeedKey(k, true);
}





int main(int argc, char *argv[]) 
{
  std::string displaymode="monoscopic";

  fprintf(stderr,"Version %s\n", VERSION_STRING(GAMEVERSION));
  fprintf(stderr,"Stormbaan Coureur is (c)2006-2008 by Bram Stolk\n");
  fprintf(stderr,"plib is (c) by Steve Baker\n");
  fprintf(stderr,"OpenDE is (c) by Russel L. Smith\n");

  char *bindirname  = dirname(argv[0]);
  if (!strcmp(bindirname,"."))
    dirprefix="/usr/share/stormbaancoureur";
  else
  {
    dirprefix = dirname(bindirname) + std::string("/share/games/stormbaancoureur");
  }
  if (getenv("PLODE_DATADIR"))
    dirprefix = getenv("PLODE_DATADIR");
  if (getenv("PLODE_DISPLAYMODE"))
    displaymode = getenv("PLODE_DISPLAYMODE");
  assert(displaymode == "monoscopic" || displaymode == "quadbufferstereoscopic" || displaymode == "passivestereoscopic");
  modelmap = new ModelMap(dirprefix);
  char *unam = getenv("USER");
  if (unam && strlen(unam)>2)
  {
    username = unam;
    if (strlen(unam)>8)
      username[8]=0;
  }
  glutInit(&argc, argv);
  int flags = GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE;
  if (displaymode == "quadbufferstereoscopic") flags = flags | GLUT_STEREO;
  glutInitDisplayMode(flags);
  glutInitWindowSize(winw, winh);
  glutInitWindowPosition(0,0);
  int winid=glutCreateWindow ("Stormbaan Coureur");
  assert(winid>=1);
  glutDisplayFunc(redraw);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);
  glutKeyboardUpFunc(keyboardUp);
  glutSpecialFunc(special);
  glutSpecialUpFunc(specialUp);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutIdleFunc(idle);
  glutFullScreen();

  do_shadows = OglCanDoShadowing();
  fprintf
  (
    stderr, 
    "This platform %s all required GL extensions to do hardware accelerated shadowing.\n",
    (do_shadows) ? "supports" : "does not support"
  );
  if (do_shadows)
  {
    OglInitShadowing();
  }

  ssgInit();

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);

  float amb[4]={0.0, 0.0, 0.0, 1};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb);

  starsky = new StarSky();
  modelmap->Put("starsky", starsky->GetEntity());

  controller_pad = new ControllerPad("/dev/input/js0");
  if (controller_pad->IsWorking())
  {
    fprintf(stderr,"Using gamepad\n");
    controller = controller_pad;
  }
  else
  {
    delete controller_pad;
    controller_pad = 0;
    char *ev = getenv("HOME");
    controller_key = new ControllerKey(std::string((ev)?ev:"~") + "/.stormbaancoureur.keys");
    controller = controller_key;
    fprintf(stderr,"Using keyboard\n");
  }


  if (displaymode != "monoscopic")
  {
    stereocontext = new StereoContext(displaymode == "quadbufferstereoscopic");
    // Move camera closer for extra wow.
    camd = 1.5;
  }
  else
  {
    monocontext = new ssgContext();
    monocontext->makeCurrent();
  }
  lightviewcontext = new ssgContext();
  float fov = spot_fov * M_PI / 180.0;
  float nearPlaneDistance = 1.0f;
  float farPlaneDistance  = 150.0f;
  float s = tan(0.5 * fov) * nearPlaneDistance;
  lightviewcontext->setFrustum(-s,s,-s,s, nearPlaneDistance, farPlaneDistance);

  ssgLight *light=ssgGetLight(0);
  const float ka = (do_shadows) ? 0.15f : 0.30f;
  const float kd = (do_shadows) ? 1.00f : 2.50f;
  const float ks = 1.0f;
  light->setColour(GL_DIFFUSE,  kd,kd,kd);
  light->setColour(GL_AMBIENT,  ka,ka,ka);
  light->setColour(GL_SPECULAR, ks,ks,ks);
  light->setPosition(0,0,8);
  light->setSpotlight(true);
  light->setSpotDirection(0,0,-1);
  light->setSpotDiffusion(100,180);
  light->setSpotAttenuation(1, 0.1, 0.04);
  light->on();

  sgCoord campos ;
  sgSetCoord (&campos, 0.0f, -13.0f,  6.0f, 0.0, -34.0f, 0.0f) ;

  bool nosound = getenv("PLODE_NO_SOUND");
  if (!nosound)
  {
    sengine = new SoundEngineAlsa(8000);
    if (!sengine->opened)
    {
      delete sengine;
      sengine = 0;
    }
  }
  if (sengine)
    intro = new Intro(sengine, dirprefix);
  gui = new PlodeGUI(username, controller_pad!=0);

  glutMainLoop();
  delete sengine;
}

